#include "Fusion_SLAM_types.h"
#include "TueObjFusn_ObjectListType.h"
#include "TueObjFusn_ErrorCodes.h"
#include "tue_prv_objectSelection_int.h"
#include "ContiRadarConverter_int.h"
#include "Fusion_SLAM_utils.h"
#include <string.h>
#include <fusion_ext.h>
#include "Sensor_LidarProcess.h"
#ifdef  RESNICS
#include "SensorLidar.h"
#endif
#include "TueObjFusn_ErrorCodes.h"
#include "tue_prv_objectSelection_int.h"

/**
 * @defgroup group_lidar_zones Lidar sensor preselection zones
 * @{
*/
#define  FUSION_PRE_OBJECTSELECTION_GLOBAL_OBSZONE_BIN         (0u)
#define  FUSION_PRV_OBJECTSELECTION_LIDARONLY_BIN              (1u)
#define  FUSION_PRV_OBJECTSELECTION_OBSZONE1_BIN               (2u)
#define  FUSION_PRV_OBJECTSELECTION_OBSZONE2_BIN               (3u)
#define  FUSION_PRV_OBJECTSELECTION_OBSZONE3_BIN               (4u)
#define  FUSION_PRV_OBJECTSELECTION_OBSZONE4_BIN               (5u)

#define  FUSION_PRV_OBJECTSELECTION_FUSED_OBSZONE1_BIN         (6u)
#define  FUSION_PRV_OBJECTSELECTION_FUSED_OBSZONE2_BIN         (7u)
#define  FUSION_PRV_OBJECTSELECTION_FUSED_OBSZONE3_BIN         (8u)
#define  FUSION_PRV_OBJECTSELECTION_FUSED_OBSZONE4_BIN         (9u)

#define  FUSION_PRV_OBJECTSELECTION_NOTCOASTED_OBSZONE1_BIN    (10u)
#define  FUSION_PRV_OBJECTSELECTION_NOTCOASTED_OBSZONE2_BIN    (11u)
#define  FUSION_PRV_OBJECTSELECTION_NOTCOASTED_OBSZONE3_BIN    (12u)
#define  FUSION_PRV_OBJECTSELECTION_NOTCOASTED_OBSZONE4_BIN    (13u)


#define  FUSION_PRV_OBJECTSELECTION_LONG_DISTANCE_ROI1_THRESHOLD_OBJ    (70.0f)
#define  FUSION_PRV_OBJECTSELECTION_LONG_DISTANCE_ROI0_THRESHOLD_OBJ    (200.0f)
#define  FUSION_PRV_OBJECTSELECTION_ANGLE_ROI2_THRESHOLD_OBJ            (10.0f)
#define  FUSION_PRV_OBJECTSELECTION_ANGLE_ROI1_THRESHOLD_OBJ            (22.5f)
#define  FUSION_PRV_OBJECTSELECTION_RANGE_THRESH_OBSZONE6               (50.0f)
#define  FUSION_PRV_OBJECTSELECTION_RANGE_THRESH_OBSZONE5               (100.0f)
#define  FUSION_PRV_OBJECTSELECTION_RANGE_THRESH_OBSZONE4               (150.0f)

/** @} */

#define FUSION_DISTANCE_TO_TRAJECTORY_THRESHOLD	4

/**
 * @def SENSOR_LIDAR2FUSION
 * 
 * @brief Defines the number of Lidar objects that are preserved
 * for the fusion process.
*/
#define SENSOR_LIDAR2FUSION		5
/**
 * @def LIDAR_SIZE
 * 
 * @brief Initial number of Lidar objects before preselection
*/
#define LIDAR_SIZE				40



extern float32 tue_prv_fusion_sqrt(float32 f32x) ;
extern float32 tue_prv_fusion_atan2(const float32 f32y, const float32 f32x) ;
extern float32 tue_prv_fusion_rtod(const float32 f32x);
extern float32 tue_prv_fusion_abs(const float32 f32x);
extern uint32 Fusion_calculateObjectRange(float32 f32PosX, float32 f32PosY, float32* f32Range);
extern uint32 Fusion_calculateObjectAngle(float32 f32PosX,float32 f32PosY, float32* f32Angle);


/**
 * @brief Calculate object range (distance)
 * 
 * Euclidean distance of object from origin is calculated
 * 
 * @param[in] f32PosX X position of track
 * @param[in] f32PosY Y position of track
 * @param[out] f32Range Object distance
 * 
 * @return error status is returned
*/
uint32 Fusion_calculateObjectRange(float32 f32PosX, float32 f32PosY, float32* f32Range)
{
	uint32 u32Success    = TUEOBJFUSN_ERROR_NOERROR;   /* PRQA S 3204 1 *//* variable may be set to other values depending on run-time pointer check activation */
	float32 f32RangeTmp;

	f32RangeTmp = ((f32PosX * f32PosX) + (f32PosY * f32PosY));
	*f32Range   = tue_prv_fusion_sqrt(f32RangeTmp);

   return u32Success;
}

/**
 * @brief Calculates object angle
 * 
 * The angle between Y and X positions is calculated in degree
 * 
 * @param[in] f32PosX X position of track
 * @param[in] f32PosY Y position of track
 * @param[out] f32Angle Object angle
 * 
 * @return error status is returned
*/
uint32 Fusion_calculateObjectAngle(float32 f32PosX,float32 f32PosY, float32* f32Angle)
{
	uint32 u32Success   = TUEOBJFUSN_ERROR_NOERROR;   /* PRQA S 3204 1 *//* variable may be set to other values depending on run-time pointer check activation */
	float32 f32AngleTmp;

//	f32PosX = pLidarObject->Kinematic.fDistX;
//	f32PosY = pLidarObject->Kinematic.fDistY;

	f32AngleTmp = tue_prv_fusion_atan2(f32PosY, f32PosX);
	f32AngleTmp = tue_prv_fusion_rtod(f32AngleTmp);
	*f32Angle = tue_prv_fusion_abs(f32AngleTmp);

   return u32Success;
}


/**
 * @brief Bin position of each track is set here
 *
 * Based on track distance from origin and the angle, it's assigned to one of
 * the 14 available zones. This function fills corresponding bin of
 * the input track. For available zones and their index, refer to @ref group_radar_zones
 * 
 * @param[in] pLidarObject Input Lidar track 
 * @param[out] as16TrkblBinCounter Array holding number of tracks in each zone (bin)
 * @param[out] fRange Track range which is calculated during the function process
 * 
 * @return Track bin index is returned.
*/
uint8 Fusion_setTrackableBinPosition(LidarObjects_t* pLidarObject, sint16 as16TrkblBinCounter[], float32* fRange)
{
	float32 f32ObjectRange;
	float32 f32ObjectAngle;
	uint32 u32Success;
	uint8 u8TrkblBinPosition = 0;
	VehDyn_t *pVehDyn	= FusionInput_RawData; 

	u32Success = Fusion_calculateObjectRange(pLidarObject->Kinematic.fDistX, pLidarObject->Kinematic.fDistY, &f32ObjectRange);
	u32Success |= Fusion_calculateObjectAngle(pLidarObject->Kinematic.fDistX, pLidarObject->Kinematic.fDistY, &f32ObjectAngle);
	BML_t_TrajRefPoint DistVEYPoint2Circle = CML_CalculateDistancePoint2Circle(
		pLidarObject->Kinematic.fDistX,
		pLidarObject->Kinematic.fDistY,
		pVehDyn->Lateral.Curve.Curve);
	float32 f_Dist2TrajAbs = BML_f_Abs(DistVEYPoint2Circle.f_DistToTraj);

	*fRange = f32ObjectRange;
	 uint8  _u16MotionType = pLidarObject->General.eObjMaintenanceState;
	
	   if (((FUSION_PRV_OBJECTSELECTION_LONG_DISTANCE_ROI1_THRESHOLD_OBJ < f32ObjectRange) &&	(FUSION_PRV_OBJECTSELECTION_ANGLE_ROI2_THRESHOLD_OBJ < f32ObjectAngle)) || 
		   ((FUSION_PRV_OBJECTSELECTION_LONG_DISTANCE_ROI1_THRESHOLD_OBJ > f32ObjectRange) &&	(FUSION_PRV_OBJECTSELECTION_ANGLE_ROI1_THRESHOLD_OBJ < f32ObjectAngle)) ||  
			(FUSION_PRV_OBJECTSELECTION_LONG_DISTANCE_ROI0_THRESHOLD_OBJ < f32ObjectRange)) 
	   {
			u8TrkblBinPosition = FUSION_PRE_OBJECTSELECTION_GLOBAL_OBSZONE_BIN;
			as16TrkblBinCounter[FUSION_PRE_OBJECTSELECTION_GLOBAL_OBSZONE_BIN]++;
	   } /* Remove all radar only stationary objects */
	   else if (FUSION_PRV_OBJECTSELECTION_RANGE_THRESH_OBSZONE4 < f32ObjectRange)
	   {
		   if (f_Dist2TrajAbs <= FUSION_DISTANCE_TO_TRAJECTORY_THRESHOLD)
		   {
				u8TrkblBinPosition = FUSION_PRV_OBJECTSELECTION_NOTCOASTED_OBSZONE1_BIN;
				as16TrkblBinCounter[FUSION_PRV_OBJECTSELECTION_NOTCOASTED_OBSZONE1_BIN]++;
		   	} else {
				u8TrkblBinPosition = FUSION_PRV_OBJECTSELECTION_FUSED_OBSZONE1_BIN;
				as16TrkblBinCounter[FUSION_PRV_OBJECTSELECTION_FUSED_OBSZONE1_BIN]++;
			}

	   }
	   else if ((FUSION_PRV_OBJECTSELECTION_RANGE_THRESH_OBSZONE4 > f32ObjectRange) && 
				(FUSION_PRV_OBJECTSELECTION_RANGE_THRESH_OBSZONE5 < f32ObjectRange))
	   {

   		   if (f_Dist2TrajAbs <= FUSION_DISTANCE_TO_TRAJECTORY_THRESHOLD)
		   {
				u8TrkblBinPosition = FUSION_PRV_OBJECTSELECTION_NOTCOASTED_OBSZONE2_BIN;
				as16TrkblBinCounter[FUSION_PRV_OBJECTSELECTION_NOTCOASTED_OBSZONE2_BIN]++;
   		   	} else {
				u8TrkblBinPosition = FUSION_PRV_OBJECTSELECTION_FUSED_OBSZONE2_BIN;
				as16TrkblBinCounter[FUSION_PRV_OBJECTSELECTION_FUSED_OBSZONE2_BIN]++;
			}

	   }
	   else if ((FUSION_PRV_OBJECTSELECTION_RANGE_THRESH_OBSZONE5 > f32ObjectRange) && 
				(FUSION_PRV_OBJECTSELECTION_RANGE_THRESH_OBSZONE6 < f32ObjectRange))
	   {
  		   if (f_Dist2TrajAbs <= FUSION_DISTANCE_TO_TRAJECTORY_THRESHOLD)
		   {
				u8TrkblBinPosition = FUSION_PRV_OBJECTSELECTION_NOTCOASTED_OBSZONE3_BIN;
				as16TrkblBinCounter[FUSION_PRV_OBJECTSELECTION_NOTCOASTED_OBSZONE3_BIN]++;			

   		   	} else {
				u8TrkblBinPosition = FUSION_PRV_OBJECTSELECTION_FUSED_OBSZONE3_BIN;
				as16TrkblBinCounter[FUSION_PRV_OBJECTSELECTION_FUSED_OBSZONE3_BIN]++;
			}
	   }
	   else if (FUSION_PRV_OBJECTSELECTION_RANGE_THRESH_OBSZONE6 > f32ObjectRange)
	   {
 		   if (f_Dist2TrajAbs <= FUSION_DISTANCE_TO_TRAJECTORY_THRESHOLD)
		   {
				 u8TrkblBinPosition = FUSION_PRV_OBJECTSELECTION_NOTCOASTED_OBSZONE4_BIN;
				 as16TrkblBinCounter[FUSION_PRV_OBJECTSELECTION_NOTCOASTED_OBSZONE4_BIN]++;
		   	} else {
				u8TrkblBinPosition = FUSION_PRV_OBJECTSELECTION_FUSED_OBSZONE4_BIN;
				as16TrkblBinCounter[FUSION_PRV_OBJECTSELECTION_FUSED_OBSZONE4_BIN]++;
			}
	   }
	   else
	   {
   			u8TrkblBinPosition = FUSION_PRE_OBJECTSELECTION_GLOBAL_OBSZONE_BIN;
			as16TrkblBinCounter[FUSION_PRE_OBJECTSELECTION_GLOBAL_OBSZONE_BIN]++;
	   }
	
	return u8TrkblBinPosition;


}

/**
 * @brief Sets which Lidar tracks must be dropped
 * 
 * This function specifies the last bin index to which tracks must be
 * deleted. It iterates through track bin indices and eliminates as many
 * track as needed so that the number of tracks to be desired.
 * 
 * @param[in] s16NumberTrackablesToDelete Number of tracks that must be deleted
 * @param[in] as16TrkblBinCounter Array specifying number of tracks in each bin
 * @param[out] u8LastBinToDelete Index of the last bin to which tracks must be deleted
 * 
 * @return Error status is returned
*/
static uint32 Fusion_setDroppedTrackableBins(uint8* u8LastBinToDelete, 
											 sint16 s16NumberTrackablesToDelete, 
											 sint16 as16TrkblBinCounter[])
{
	uint32 u32Success = TUEOBJFUSN_ERROR_NOERROR;   /* PRQA S 3204 1 *//* variable may be set to other values depending on run-time check activation */
	uint8 u8Bin;
	uint8 u8LastBin;


	u8LastBin = TUE_PRV_OBJECTSELECTION_MAX_BINS;

	for (u8Bin = 0u; (u8Bin < TUE_PRV_OBJECTSELECTION_MAX_BINS) && (0 < s16NumberTrackablesToDelete); u8Bin++)
	{
	  if (as16TrkblBinCounter[u8Bin] < s16NumberTrackablesToDelete)
	  {  
	     /* bin has to be dropped completly -> sets bin to be dropped (TUE_PRV_OBJECTSELECTION_DROP_BIN)*/
	     s16NumberTrackablesToDelete -= as16TrkblBinCounter[u8Bin];
	  }
	  else
	  {  
	     /* bin has to be dropped party -> write number which has to be dropped */ 
	     s16NumberTrackablesToDelete = 0;
	     u8LastBin = u8Bin; /* this is the last bin which has to be (partly) deleted */
	  }
	}
   *u8LastBinToDelete = u8LastBin;


   return u32Success;
}

/**
 * @brief Copies preserved Lidar tracks and converts the type
 * 
 * A boolean list determining which elements must be dropped is 
 * provided as input and this function only keeps a subset of tracks
 * based on the array. Also the type of each track is converted to fusion 
 * core type.
 * 
 * @param[in] pLidarObjList original Lidar object list
 * @param[in] abObjectsToBeDropped array showing which tracks must be deleted
 * @param[out] pTueObjList converted Lidar object list
 * 
 * @return error status is returned
*/
uint32 Fusion_copyLkfTrkbl2Outputlist(LidarObjectList_t* pLidarObjList, 
									  const boolean abObjectsToBeDropped[], 
									  TueObjFusn_ObjectListType* pTueObjList)
{
	uint32 u32Success = TUEOBJFUSN_ERROR_NOERROR;
	uint16 u16i;
	uint16 u16NumObj      = 0u;
	sint16 s16Idx;

	LidarObjects_t* pLidarObject;
	TueObjFusn_TrackableType* pFusnObject;

	for (u16i = 0u; u16i < BSW_OBJECT_NUM_MAX && u16NumObj < SENSOR_LIDAR2FUSION; u16i++){
		pLidarObject = &(pLidarObjList->Objects[u16i]);
		if ((FALSE == abObjectsToBeDropped[u16i]) && (pLidarObject->ObjectId != 0xFF)){
			pFusnObject = &(pTueObjList->aTrackable[u16NumObj]);
			// TODO: change radar to lidar accordingly 
			ContiRaObjToTrackableConverter(pFusnObject, pLidarObject);
			//LogWrite("ContiRaObjToTrackableConverter %d, %d, %.02f,  %.02f",
			// u16i, u16NumObj,
			// pFusnObject->vecX.data[0],pFusnObject->vecX.data[1]);
			u16NumObj += 1u;
		}
		else{
			/* MISRA */
		}

	}  /* end for u16i */
   pTueObjList->u16NumObjects = u16NumObj;

   return u32Success;
}



/** 
 * @brief Lidar object preselection
 *
 * Specifes which objects to be preserved and which to be 
 * thrown away.
 * 
 * @param[in] pLidarObjList list of all Lidar objects
 * @param[out] abObjectsToBeDropped boolean list determining which objects to throw away
 *
 * @return
 *     none
*/
void Fusion_prv_object_select(LidarObjectList_t* pLidarObjList, boolean abObjectsToBeDropped[])
{
	uint32 u32Success;
	uint8 u8LastBinToDelete;
	uint16 u16i, u8Bin;
	uint8 i, j;
	LidarObjects_t *pLidarObject;
	uint8 auBinPosition[BSW_OBJECT_NUM_MAX]= {0};
	sint16 as16TrkblBinCounter[TUE_PRV_OBJECTSELECTION_MAX_BINS] = {0};
	float32 Dist2TrajAbsArray[BSW_OBJECT_NUM_MAX] = {0};
	float32 f32ObjectRange;
	sint16 s16ObjectsToDelete = 0;
	if(pLidarObjList->HeaderObjList.iNumOfUsedObjects > SENSOR_LIDAR2FUSION) { 
		s16ObjectsToDelete = pLidarObjList->HeaderObjList.iNumOfUsedObjects - SENSOR_LIDAR2FUSION;
	}


	for(u16i = 0u; u16i < BSW_OBJECT_NUM_MAX; u16i++)
	{
	   abObjectsToBeDropped[u16i] = FALSE;
	}

	// pLidarObjList->HeaderObjList.iNumOfUsedObjects
	for (u16i = 0u; u16i < BSW_OBJECT_NUM_MAX; u16i++)
	{
	   pLidarObject = &pLidarObjList->Objects[u16i];
	   auBinPosition[u16i] = 0xFF;
	   if(pLidarObject->ObjectId != 0xFF) {
		   auBinPosition[u16i] = Fusion_setTrackableBinPosition(pLidarObject, as16TrkblBinCounter, &f32ObjectRange);
		   Dist2TrajAbsArray[u16i] = f32ObjectRange;
	   }
	}
	for (u16i = 0u; u16i < TUE_PRV_OBJECTSELECTION_MAX_BINS; u16i++)
	{
		//LogWrite("Fusion_setTrackableBinPosition: %d, %d",
								// u16i,
								// as16TrkblBinCounter[u16i]);

	}


	if(as16TrkblBinCounter[0] > s16ObjectsToDelete) {
		s16ObjectsToDelete = as16TrkblBinCounter[0];
	}

	Fusion_setDroppedTrackableBins(&u8LastBinToDelete, s16ObjectsToDelete, as16TrkblBinCounter);
//	LogWrite("Fusion_prv_object_select  :%d, %d, %d",
//							pExtObjList->HeaderObjList.iNumOfUsedObjects,
//							s16ObjectsToDelete,
//							u8LastBinToDelete);

	/* loop through bins and delete trackables */
	for (u8Bin = 0u; u8Bin <= u8LastBinToDelete; u8Bin++) 
	{ 
	   /* loop through all trackables and delete which are in this bin */
		if(s16ObjectsToDelete >= as16TrkblBinCounter[u8Bin]) {
		   for (u16i = 0u; (u16i < BSW_OBJECT_NUM_MAX) && (0 < s16ObjectsToDelete); u16i++) 
		   {
			  /* Remove coasted stationary objects */
			  if (u8Bin == auBinPosition[u16i])  
			  {
				 s16ObjectsToDelete--;
				 abObjectsToBeDropped[u16i] = TRUE;
				 //LogWrite("Fusion_prv_object_select 1 abObjectsToBeDropped :%d", u16i);
			  }   
			  else
			  {
				 /* MISRA */
			  }
		   }
		} else {
			break;
		}
	}
	
	if(s16ObjectsToDelete > 0) {
		sint8 auBinIdx[BSW_OBJECT_NUM_MAX]= {0};
		uint8 sobjNum = 0;
		sint8 arrayIndex = -1;
		for(i = 0; i < BSW_OBJECT_NUM_MAX; i++) {
			if(auBinPosition[i] == u8Bin) {
				auBinIdx[sobjNum] = i;
				sobjNum++;
			}
		}
	
		// Removes s16ObjectsToDelete number of furthest tracks
		for(i = 0; i < s16ObjectsToDelete; i++) {
			float32 tempMinDist = -1; // This variable finds largest distance of a track (Min?)
			for(j = 0; j < sobjNum; j++) {
				if (Dist2TrajAbsArray[auBinIdx[j]] >= 0)
				{
					if (BML_f_Abs(tempMinDist + 1) < FLT_EPSILON || Dist2TrajAbsArray[auBinIdx[j]] > tempMinDist)
					{
						tempMinDist = Dist2TrajAbsArray[auBinIdx[j]];
						arrayIndex = auBinIdx[j];
					}
				}

			}
			if (arrayIndex >= 0 && arrayIndex < BSW_OBJECT_NUM_MAX )
			{
				Dist2TrajAbsArray[arrayIndex] = -1;
				abObjectsToBeDropped[arrayIndex] = TRUE;
				//LogWrite("Fusion_prv_object_select 2 abObjectsToBeDropped :%d", arrayIndex);
			} else {
				LogWrite("Fusion_prv_object_select ERROR ");

			}
		}

	}

}

/** 
 * @brief Selects a subset of Lidar objects to be processed and converts types
 *
 * Only a subset of Lidar objects are selected to be processed and fused. 
 * Specifically, #SENSOR_LIDAR2FUSION defines how many to keep.
 * 
 * @param[in] pLidarObjList list of all Lidar objects
 * @param[in] u32globalTimeStamp global system timestamp
 * @param[out] pTueObjList list of preselected Lidar objects
 *
 * @return none
******************************************************************************/
void Fusion_LidarObjectPreSelection(LidarObjectList_t* pLidarObjList, 
									TueObjFusn_ObjectListType *pTueObjList, 
									uint32 u32globalTimeStamp)
{
	boolean abObjectsToBeDropped[BSW_OBJECT_NUM_MAX] = {0};
	float32 f32CorrLatency	= FLT_ZERO;
//	boolean abProcessTracks[CONTI_LIDAR_CONVERTER_OBJECTS_U8NUMOBJECTS_MAX];
	
	/* PRQA S 3223 1 *//* static variable requires function-scope only */
	static uint16 u16updateCount = 0u;
	/* PRQA S 3223 1 *//* static variable requires function-scope only */
	static uint32 u32radarTimeStampPrev = 0u;


	if(u32radarTimeStampPrev != pLidarObjList->sSigHeader.uiTimeStamp)
	{
	   u16updateCount = ((u16updateCount + 1u) % TUEOBJFUSN_OBJECTLIST_U16LISTUPDATECOUNTER_MAX);
	   u32radarTimeStampPrev = (pLidarObjList->sSigHeader.uiTimeStamp);
	}
	else
	{
	   /* MISRA */
	}

	pTueObjList->u16ListUpdateCounter	= u16updateCount;
	
	/* correct latency using global timestamp */
	if (u32globalTimeStamp >= u32radarTimeStampPrev)
	{
		f32CorrLatency = (float32)(u32globalTimeStamp - u32radarTimeStampPrev);
	}
	else
	{
		f32CorrLatency = (float32)(0xFFFFFFFF - u32radarTimeStampPrev + u32globalTimeStamp);
	}
	f32CorrLatency = MAX(0, MIN(1000000, (f32CorrLatency)));
	pTueObjList->f32MeasurementLatency  = (f32CorrLatency) / 1000.0 / 1000.0 + 0.02; //convert latency from us to s 
	pTueObjList->u32SensorPattern		 = TUEOBJFUSN_TRACKABLE_U32SENSOR_LIDAR_FRONT;

	//LogWrite("Fusion_LidarObjectPreSelection u16updateCount :%d, %d", u16updateCount, pExtObjList->HeaderObjList.iNumOfUsedObjects);

	Fusion_prv_object_select(pLidarObjList, abObjectsToBeDropped);
	Fusion_copyLkfTrkbl2Outputlist(pLidarObjList, abObjectsToBeDropped, pTueObjList);

}

